using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Web.UI.WebControls;

using CMS.Base.Web.UI;
using CMS.Search;
using CMS.Search.Azure;
using CMS.PortalEngine.Web.UI;

using Microsoft.Azure.Search;
using Microsoft.Azure.Search.Models;
using CMS.Helpers;
using System.Web.UI;

public partial class CMSWebParts_Custom_CustomAzureSearch : CMSAbstractWebPart
{
    // The SearchIndexClient instance used to interact with the specified Azure Search index
    private ISearchIndexClient searchIndexClient = InitializeIndex("dg-store");

    // The fields used for faceted navigation
    private const string FACET_COFFEE_COUNTRY = "coffeecountry";
    private const string FACET_COFFEE_FARM = "coffeefarm";
    public readonly List<string> Facets = new List<string>
    {
        FACET_COFFEE_COUNTRY,
        FACET_COFFEE_FARM
    };

    #region Properties

    public string HighlightFields
    {
        get
        {
            return ValidationHelper.GetString(GetValue("HighlightFields"), "");
        }
        set
        {
            SetValue("HighlightFields", value);
        }
    }

    public string HighlightHTMLPrefix
    {
        get
        {
            return ValidationHelper.GetString(GetValue("HighlightHTMLPrefix"), "");
        }
        set
        {
            SetValue("HighlightHTMLPrefix", value);
        }
    }


    public string HighlightHTMLSuffix
    {
        get
        {
            return ValidationHelper.GetString(GetValue("HighlightHTMLSuffix"), "");
        }
        set
        {
            SetValue("HighlightHTMLSuffix", value);
        }
    }

    #endregion


    /// <summary>
    /// Returns an initialized SearchServiceClient instance based on the specified index code name.
    /// </summary>
    private static ISearchIndexClient InitializeIndex(string indexName)
    {
        // Converts the Kentico index code name to a valid Azure Search index name (if necessary)
        indexName = NamingHelper.GetValidIndexName(indexName);
        SearchIndexInfo index = SearchIndexInfoProvider.GetSearchIndexInfo(indexName);
        SearchServiceClient client = new SearchServiceClient(index.IndexSearchServiceName, new SearchCredentials(index.IndexQueryKey));
        return client.Indexes.GetClient(indexName);
    }

    protected void Page_Load(object sender, EventArgs e)
    {
        try
        {
            // Prepares a list of filter queries for the search request
            var filterQueries = new List<string>()
        {
            // Filters the search results to display only pages (products) of the 'dancinggoat.coffee' type
            "classname eq 'dancinggoat.coffee'"
        };

            // Adds filter queries based on the selected options in the faceted navigation (coffee farm and region)
            var selectedCountries = filterCountry.GetSelectedItems();
            var selectedFarms = filterFarm.GetSelectedItems();

            if (selectedCountries.Any())
            {
                filterQueries.Add(GetFilterQuery(selectedCountries, FACET_COFFEE_COUNTRY));
            }
            if (selectedFarms.Any())
            {
                filterQueries.Add(GetFilterQuery(selectedFarms, FACET_COFFEE_FARM));
            }

            // Prepares the search parameters
            var searchParams = new Microsoft.Azure.Search.Models.SearchParameters
            {
                Facets = Facets,
                Filter = String.Join(" and ", filterQueries)
            };

            if (cbShowHighlights.Checked)
            {
                if (this.HighlightFields != "")
                {
                    searchParams.HighlightFields = this.HighlightFields.Split(',').ToList();

                    searchParams.HighlightPreTag = this.HighlightHTMLPrefix;
                    searchParams.HighlightPostTag = this.HighlightHTMLSuffix;
                }
            }

            // Check if a scoring profile was selected
            if (ddlScoringProfile.SelectedValue != "")
            {
                searchParams.ScoringProfile = ddlScoringProfile.SelectedValue;
            }

            // Gets the search text from the input
            string searchString = txtSearch.Text + "*";

            // Performs the search request for the specified Azure Search index and parameters
            DocumentSearchResult result = searchIndexClient.Documents.Search(searchString, searchParams);

            // Fills or updates the faceted navigation options
            if (result.Facets != null)
            {
                if (!IsPostBack)
                {
                    foreach (var resultFacet in result.Facets)
                    {
                        foreach (FacetResult value in resultFacet.Value)
                        {
                            AddItemToCheckBoxList(value, resultFacet.Key);
                        }
                    }
                }
                else
                {
                    foreach (ListItem item in filterCountry.Items)
                    {
                        UpdateCountListItem(result.Facets[FACET_COFFEE_COUNTRY], item);
                    }

                    foreach (ListItem item in filterFarm.Items)
                    {
                        UpdateCountListItem(result.Facets[FACET_COFFEE_FARM], item);
                    }
                }
            }

            // Displays the search results
            GenerateResultView(result);
        }
        catch(Exception ex)
        {
            ShowMessage(MessageTypeEnum.Error, "There was an error processing the request.", ex.Message, null, true);
        }
    }

    /// <summary>
    /// Builds a filter query based on selected faceted navigation options.
    /// </summary>
    private string GetFilterQuery(IEnumerable<ListItem> selectedItems, string column)
    {
        var queries = selectedItems.Select(item => $"{column} eq '{item.Value.Replace("'", "''")}'");
        return $"({String.Join(" or ", queries)})";
    }

    /// <summary>
    /// Adds a retrieved FacetResult to the 'listItem' filtering options.
    /// </summary>
    private void AddItemToCheckBoxList(FacetResult facetResult, string resultFacetKey)
    {
        var item = new ListItem($"{facetResult.Value} ({facetResult.Count})", facetResult.Value.ToString());
        switch (resultFacetKey)
        {
            case FACET_COFFEE_COUNTRY:
                filterCountry.Items.Add(item);
                break;
            case FACET_COFFEE_FARM:
                filterFarm.Items.Add(item);
                break;
            default:
                break;
        }
    }

    /// <summary>
    /// Updates the counts of matching results for the 'listItem' filtering options according to the retrieved facet data.
    /// </summary>
    private static void UpdateCountListItem(IEnumerable<FacetResult> facetResults, ListItem listItem)
    {
        long? count = 0;
        foreach (var items in facetResults)
        {
            if (items.Value.Equals(listItem.Value))
            {
                count = items.Count;
                break;
            }
        }

        listItem.Text = $"{listItem.Value} ({count})";
    }

    /// <summary>
    /// Displays the search results filtered according to the selected parameters.
    /// </summary>
    private void GenerateResultView(DocumentSearchResult searchResult)
    {
        if (searchResult.Results.Count == 0)
        {
            lblSearchResults.Text = "No search results found.";
            return;
        }

        var resultText = new StringBuilder();

        foreach (SearchResult result in searchResult.Results)
        {
            resultText.Append($"<div>");

            // Displays the 'skushortdescription' field's value for result items without highlights matching the search text
            if (result.Highlights == null)
            {
                resultText.Append("<h3><a href='" + URLHelper.ResolveUrl(ValidationHelper.GetString(result.Document["nodealiaspath"], "/")) + "'>" + ValidationHelper.GetString(result.Document["skuname"], "") + "</a></h3>");
                resultText.Append("<p>" + ValidationHelper.GetString(result.Document["skushortdescription"], "") + "</p>");
                resultText.Append("<p><strong>Country:</strong> " + ValidationHelper.GetString(result.Document["coffeecountry"], "") + "</p>");
                resultText.Append("<p><strong>Farm:</strong> " + ValidationHelper.GetString(result.Document["coffeefarm"], "") + "</p>");
            }
            // Displays highlights matching the search text from the 'HighlightFields' specified in the SearchParameters
            else
            {
                if (result.Highlights.ContainsKey("skuname"))
                {
                    resultText.Append("<h3><a href='" + URLHelper.ResolveUrl(ValidationHelper.GetString(result.Document["nodealiaspath"], "/")) + "'>" + result.Highlights["skuname"][0].ToString() + "</a></h3>");
                }
                else
                {
                    resultText.Append("<h3><a href='" + URLHelper.ResolveUrl(ValidationHelper.GetString(result.Document["nodealiaspath"], "/")) + "'>" + result.Document["skuname"] + "</a></h3>");
                }

                if (result.Highlights.ContainsKey("skushortdescription"))
                {
                    resultText.Append("<p>" + ValidationHelper.GetString(result.Highlights["skushortdescription"][0], "") + "</p>");
                }
                else
                {
                    resultText.Append("<p>" + ValidationHelper.GetString(result.Document["skushortdescription"], "") + "</p>");
                }
                
                if (result.Highlights.ContainsKey("coffeecountry"))
                {
                    resultText.Append("<p><strong>Country:</strong> " + ValidationHelper.GetString(result.Highlights["coffeecountry"][0], "") + "</p>");
                }
                else
                {
                    resultText.Append("<p><strong>Country:</strong> " + ValidationHelper.GetString(result.Document["coffeecountry"], "") + "</strong></p>");
                }
                
                if (result.Highlights.ContainsKey("coffeefarm"))
                {
                    resultText.Append("<p><strong>Farm:</strong> " + ValidationHelper.GetString(result.Highlights["coffeefarm"][0], "") + "</p>");
                }
                else
                {
                    resultText.Append("<p><strong>Farm:</strong> " + ValidationHelper.GetString(result.Document["coffeefarm"], "") + "</strong></p>");
                }
            }

            if (cbShowScore.Checked)
            {
                resultText.Append("<p>Score: " + result.Score + "</p>");
            }
            resultText.Append($"</div>");
        }

        lblSearchResults.Text = resultText.ToString();
    }

    protected void txtSearch_TextChanged(object sender, EventArgs e)
    {
        pnlSuggestions.Visible = false;
        lblSuggestions.Text = "";

        // Check if there are any suggestions
        if (!String.IsNullOrWhiteSpace(txtSearch.Text))
        {
            if (txtSearch.Text.Length > 2)
            {
                // Retrieves suggestions based on search input using 'productnamesuggester'
                SuggestParameters sp = new SuggestParameters();
                sp.UseFuzzyMatching = true;
                sp.Top = 3;
                DocumentSuggestResult suggestResult = searchIndexClient.Documents.Suggest(txtSearch.Text, "productnamesuggester", sp);
                if(suggestResult.Results.Count > 0)
                {
                    foreach(SuggestResult result in suggestResult.Results)
                    {
                        LinkButton lb = new LinkButton();
                        lb.Text = result.Text;
                        lb.OnClientClick = "return setSearch('" + result.Text + "');";
                        pnlSuggestions.Controls.Add(lb);
                        pnlSuggestions.Controls.Add(new LiteralControl("<br />"));
                    }
                    pnlSuggestions.Visible = true;
                }
            }
        }
    }

    protected void btnReset_Click(object sender, EventArgs e)
    {
        txtSearch.Text = "";
        cbShowScore.Checked = false;
        cbShowHighlights.Checked = false;
        Page_Load(null, null);
    }
}